import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

CircularArray = class();

function CircularArray:Constructor()
	self.data = { };
	self.data.length = 0;
end

function CircularArray:realIndex(index)
	if self.data.length == 0 then
		return 0;
	end
	local x = index % self.data.length;
	return x;
end

function CircularArray:get(index)
	return self.data[self:realIndex(index)];
end

function CircularArray:insert(index, item, forward)
	index = self:realIndex(index);

	if forward then
		if self.data.length - 1 >= index then
			for i = self.data.length - 1, index, -1 do
				self.data[i + 1] = self.data[i];
			end
		end
	else
		self.data[self.data.length] = self.data[0];
		for i = 0, index - 1, 1 do
			self.data[i] = self.data[i + 1];
		end
	end

	self.data[index] = item;
	self.data.length = self.data.length + 1;
	return index;
end

function CircularArray:remove(index)
	index = self:realIndex(index);

	if self.data.length > 0 then
		for i = index, self.data.length - 2, 1 do
			self.data[i] = self.data[i + 1];
		end
		self.data[self.data.length - 1] = nil;
		self.data.length = self.data.length - 1;
	end

	return index;
end

CircularQuickslot = class(Turbine.UI.Lotro.Quickslot);

function CircularQuickslot:MouseClick(args)
	self:GetParent():toggle();
end

function CircularQuickslot:MouseEnter(args)
	self.state = 0;
end

function CircularQuickslot:MouseLeave(args)
	self.state = 1;
end

function CircularQuickslot:MouseUp(args)
	if self.state == 1 then
		self.state = 2;
	end

	self:GetParent():redraw();
end

function CircularQuickslot:DragLeave(args)
	if self.state == 2 then
		self:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
	end

	self:GetParent():redraw();
	self.state = 0;
end

function CircularQuickslot:MouseWheel(args)
	local parent = self:GetParent();

	parent.settings.index = parent.array:realIndex(parent.settings.index - tonumber(args.Direction));
	parent:redraw();
end

function CircularQuickslot:ShortcutChanged(args)
	if self.safe then
		return;
	end

	local shortcut = self:GetShortcut();
	local parent = self:GetParent();

	if shortcut == nil or shortcut:GetType() == Turbine.UI.Lotro.ShortcutType.Undefined then
		local index = parent.array:remove(parent.settings.index + self.index);
		if index < parent.settings.index then
			parent.settings.index = parent.array:realIndex(parent.settings.index - 1);
		end
	else
		local item = { };

		item.type = shortcut:GetType();
		item.data = shortcut:GetData();
		local index = parent.array:insert(parent.settings.index + self.index, item, self.index >= 0);
		if self.index > 0 then
			if index <= parent.settings.index then
				parent.settings.index = parent.array:realIndex(parent.settings.index + 1);
			end
		elseif self.index < 0 then
			if index >= parent.settings.index then
				parent.settings.index = parent.array:realIndex(parent.settings.index - 1);
			end
		end
	end

	parent:redraw();
	parent:saveData();
end

function CircularQuickslot:Constructor(parent, size)
	Turbine.UI.Lotro.Quickslot.Constructor(self);

	self:SetParent(parent);
	self:SetSize(size, size);

	self.state = 0;
	self.safe = false;
end

function CircularQuickslot:SafeSetShortcut(shortcut)
	self.safe = true;
	local status = pcall(self.SetShortcut, self, shortcut);
	self.safe = false;
	return status;
end

CircularQuickslotBar = class(Turbine.UI.Window);

function CircularQuickslotBar:Constructor(name, count, settings, data)
	Turbine.UI.Window.Constructor(self);

	self.hide = false
	self.visible = true;
	self.name = name;
	self.spacing = 0;
	self.count = 0;
	self.quickslotCount = 0;
	self.quickslotSize = 36;
	self.quickslots = { };
	self.array = CircularArray();
	if settings ~= nil then
		self.settings = settings;
	else
		self.settings = { };
		self.settings.count = 3;
		self.settings.index = 0;
	end
	if data ~= nil then
		self.array.data = data;
	end

	if count ~= nil then
		self.settings.count = count;
	end

	self:setCount(self.settings.count);
	self:SetOpacity(1);
	self:SetWantsKeyEvents(true);

	self:SetVisible();
end

function CircularQuickslotBar:KeyDown(args)
	if args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function CircularQuickslotBar:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end
	Turbine.UI.Window.SetVisible(self, self.visible and not self.hide);
end

function CircularQuickslotBar:IsVisible()
	return self.visible;
end

function CircularQuickslotBarLoadData(name)
	data = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_data_" .. name);

	if type(data) ~= "table" then
		return nil;
	end

	return data;
end

function CircularQuickslotBar:saveData()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_data_" .. self.name, self.array.data);
end


function CircularQuickslotBarLoadSettings(name)
	settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_settings_" .. name);
	if type(settings) ~= "table" then
		return nil;
	end

	if not settings.count then
		settings.count = 3;
	end

	if not settings.index then
		settings.index = 0;
	end

	return settings;
end

function CircularQuickslotBar:saveSettings()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_settings_" .. self.name, self.settings);
end


function CircularQuickslotBar:reposition()
	local x, y = Turbine.UI.Display:GetMousePosition();
	local width = self.quickslotSize;
	local height = (self.quickslotSize + self.spacing) * self.count - self.spacing;

	self:SetPosition(x - (width / 2), y - (height / 2));
end

function CircularQuickslotBar:redraw()
	local width = self.quickslotSize;
	local height = (self.quickslotSize + self.spacing) * self.count - self.spacing;

	for i = 0, self.count - 1, 1 do
		local item = self.array:get(self.quickslots[i].index + self.settings.index);
		if item == nil then
			self.quickslots[i]:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
			self.quickslots[i]:SetVisible(true);
		else
			local shortcut = Turbine.UI.Lotro.Shortcut(item.type, item.data);
			if not self.quickslots[i]:SafeSetShortcut(shortcut) then
				i = i - 1;
				self.array:remove(self.quickslots[i].index + self.settings.index);
			end
			self.quickslots[i]:SetVisible(true);
		end
	end

	self:SetSize(width + 1, height + 1);
	self:SetSize(width, height);
end

function CircularQuickslotBar:setCount(newCount)
	if self.count < newCount then
		for i = self.count, newCount - 1, 1 do
			self.quickslots[i] = CircularQuickslot(self, self.quickslotSize);
		end
	elseif self.count > newCount then
		for i = self.count - 1, newCount, -1 do
			self.quickslots[i]:SetVisible(false);
			self.quickslots[i]:SetParent(nil);
			self.quickslots[i] = nil;
		end
	end

	for i = 0, newCount - 1, 1 do
		self.quickslots[i].index = math.ceil(-newCount / 2) + i;
		self.quickslots[i]:SetPosition(0, i * (self.quickslotSize + self.spacing));
	end

	local width = self.quickslotSize;
	local height = (self.quickslotSize + self.spacing) * newCount + self.spacing

	self:SetSize(width, height);

	self.count = newCount;
	self.settings.count = newCount;

	self:reposition();
	self:saveSettings();
end

function CircularQuickslotBar:toggle()
	if self:IsVisible() then
		self:SetVisible(false);
	else
		self:reposition();
		self:redraw();
		self:SetVisible(true);
		self:Focus();
	end
end

function CircularQuickslotBar:random()
	self.settings.index = math.random(self.array.data.length) - 1;
end

function Turbine.Plugin:Unload(sender, args)
	local i = 0;
	local names = { };

	for k, v in pairs(CircularQuickslotBars) do
		CircularQuickslotBars[k]:saveSettings();
		CircularQuickslotBars[k]:saveData();
		names[i] = k;
		i = i + 1;
	end

	names.length = i;
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_names", names);
end


CircularQuickslotCommand = class(Turbine.ShellCommand);

function CircularQuickslotCommand:Execute(command, arguments)
	if arguments ~= "" then
		local i = string.find(arguments, " ");
		local name = arguments;
		local count = nil;
		local r = false;

		if i ~= nil then
			local s = string.sub(arguments, i + 1, -1)
			name = string.sub(arguments, 0, i - 1);

			if s == "random" then
				r = true;
			else
				count = tonumber(s);
			end
		end

		if CircularQuickslotBars[name] == nil then
			CircularQuickslotBars[name] = CircularQuickslotBar(name, count, nil, nil);

			local j = 0;
			local names = { };

			for k, v in pairs(CircularQuickslotBars) do
				names[j] = k;
				j = j + 1;
			end
			names.length = j;
			Turbine.PluginData.Save(Turbine.DataScope.Character, "CircularQuickbar_names", names);
		else
			if count ~= nil then
				CircularQuickslotBars[name]:setCount(count);
			end
			if r then
				CircularQuickslotBars[name]:random();
			end
			CircularQuickslotBars[name]:toggle();
		end
	else
		Turbine.Shell.WriteLine("Usage: /cq <name> [#]");
	end
end

function CircularQuickslotCommand:GetShortHelp()
	return "Circular quickslot";
end

circularQuickslotCommand = CircularQuickslotCommand();
Turbine.Shell.AddCommand('circularquickslot', circularQuickslotCommand);
Turbine.Shell.AddCommand('cq', circularQuickslotCommand);

CircularQuickslotBars = { };

names = Turbine.PluginData.Load(Turbine.DataScope.Character, "CircularQuickbar_names");
if names ~= nil then
	for i = 0, names.length - 1, 1 do
		local data = CircularQuickslotBarLoadData(names[i]);
		local settings = CircularQuickslotBarLoadSettings(names[i]);
		if data ~= nil and settings ~= nil then
			CircularQuickslotBars[names[i]] = CircularQuickslotBar(names[i], nil, settings, data);
			CircularQuickslotBars[names[i]]:toggle();
		end
	end
end

math.randomseed(Turbine.Engine.GetGameTime());